from os import getpid

from dask.bag import from_sequence
import dask.config
from dask.distributed import Client
from eliot import log_call, to_file
from eliot.dask import compute_with_trace


@log_call
def multiply(x, y=7):
    return x * y

@log_call
def add(x, y):
    return x + y

@log_call
def main_computation():
    bag = from_sequence([1, 2, 3])
    bag = bag.map(multiply).fold(add)
    return compute_with_trace(bag)[0]  # instead of dask.compute(bag)

def _start_logging():
    # Name log file based on PID, so different processes so stomp on each
    # others' logfiles:
    to_file(open("{}.log".format(getpid()), "a"))

def main():
    # Setup logging on the main process:
    _start_logging()

    # Start three worker processes on the local machine:
    client = Client(n_workers=3, threads_per_worker=1)

    # Setup Eliot logging on each worker process:
    client.run(_start_logging)

    # Run the Dask computation in the worker processes:
    result = main_computation()
    print("Result:", result)


if __name__ == '__main__':
    import dask_eliot
    dask_eliot.main()



"""
Tests for coroutines, for Python versions that support them.
"""

import sys
if sys.version_info[:2] >= (3, 5):
    from .corotests import CoroutineTests, ContextTests


__all__ = ["CoroutineTests", "ContextTests"]



"""
Tests for L{eliot.json}.
"""

from __future__ import unicode_literals, absolute_import

from unittest import TestCase, skipUnless, skipIf
from json import loads, dumps

try:
    import numpy as np
except ImportError:
    np = None

from eliot.json import EliotJSONEncoder


class EliotJSONEncoderTests(TestCase):
    """Tests for L{EliotJSONEncoder}."""

    @skipUnless(np, "NumPy not installed.")
    def test_numpy(self):
        """NumPy objects get serialized to readable JSON."""
        l = [np.float32(12.5), np.float64(2.0), np.float16(0.5),
             np.bool(True), np.bool(False), np.bool_(True),
             np.unicode_("hello"),
             np.byte(12), np.short(12), np.intc(-13), np.int_(0),
             np.longlong(100), np.intp(7),
             np.ubyte(12), np.ushort(12), np.uintc(13),
             np.ulonglong(100), np.uintp(7),
             np.int8(1), np.int16(3), np.int32(4), np.int64(5),
             np.uint8(1), np.uint16(3), np.uint32(4), np.uint64(5)]
        l2 = [l, np.array([1, 2, 3])]
        roundtripped = loads(dumps(l2, cls=EliotJSONEncoder))
        self.assertEqual([l, [1, 2, 3]], roundtripped)

    @skipIf(np, "NumPy is installed.")
    def test_numpy_not_imported(self):
        """If NumPy is not available, EliotJSONEncoder continues to work.

        This ensures NumPy isn't a hard dependency.
        """
        with self.assertRaises(TypeError):
            dumps([object()], cls=EliotJSONEncoder)
        self.assertEqual(dumps(12, cls=EliotJSONEncoder), "12")



"""
Tests for L{eliot._generators}.
"""

from __future__ import unicode_literals, absolute_import

from pprint import pformat
from unittest import TestCase

from eliot import (
    Message,
    start_action,
)
from ..testing import (
    capture_logging,
    assertHasAction,
)

from .._generators import (
    use_generator_context,
    eliot_friendly_generator_function,
)
from .._action import _context_owner


def assert_expected_action_tree(testcase, logger, expected_action_type, expected_type_tree):
    """
    Assert that a logger has a certain logged action with certain children.

    @see: L{assert_generator_logs_action_tree}
    """
    logged_action = assertHasAction(
        testcase,
        logger,
        expected_action_type,
        True,
    )
    type_tree = logged_action.type_tree()
    testcase.assertEqual(
        {expected_action_type: expected_type_tree},
        type_tree,
        "Logger had messages:\n{}".format(pformat(logger.messages, indent=4)),
    )


def assert_generator_logs_action_tree(testcase, generator_function, logger, expected_action_type, expected_type_tree):
    """
    Assert that exhausting a generator from the given function logs an action
    of the given type with children matching the given type tree.

    @param testcase: A test case instance to use to make assertions.
    @type testcase: L{unittest.TestCase}

    @param generator_function: A no-argument callable that returns a generator
        to be exhausted.

    @param logger: A logger to inspect for logged messages.
    @type logger: L{MemoryLogger}

    @param expected_action_type: An action type which should be logged by the
        generator.
    @type expected_action_type: L{unicode}

    @param expected_type_tree: The types of actions and messages which should
        be logged beneath the expected action.  The structure of this value
        matches the structure returned by L{LoggedAction.type_tree}.
    @type expected_type_tree: L{list}
    """
    list(eliot_friendly_generator_function(generator_function)())
    assert_expected_action_tree(
        testcase,
        logger,
        expected_action_type,
        expected_type_tree,
    )


class EliotFriendlyGeneratorFunctionTests(TestCase):
    """
    Tests for L{eliot_friendly_generator_function}.
    """
    # Get our custom assertion failure messages *and* the standard ones.
    longMessage = True

    def setUp(self):
        use_generator_context()
        self.addCleanup(_context_owner.reset)

    @capture_logging(None)
    def test_yield_none(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type=u"hello")
            yield
            Message.log(message_type=u"goodbye")
        g.debug = True  # output yielded messages

        with start_action(action_type=u"the-action"):
            list(g())

        assert_expected_action_tree(
            self,
            logger,
            u"the-action",
            [u"hello", u"yielded", u"goodbye"],
        )

    @capture_logging(None)
    def test_yield_value(self, logger):
        expected = object()

        @eliot_friendly_generator_function
        def g():
            Message.log(message_type=u"hello")
            yield expected
            Message.log(message_type=u"goodbye")
        g.debug = True  # output yielded messages

        with start_action(action_type=u"the-action"):
            self.assertEqual([expected], list(g()))

        assert_expected_action_tree(
            self,
            logger,
            u"the-action",
            [u"hello", u"yielded", u"goodbye"],
        )

    @capture_logging(None)
    def test_yield_inside_another_action(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type=u"a")
            with start_action(action_type=u"confounding-factor"):
                Message.log(message_type=u"b")
                yield None
                Message.log(message_type=u"c")
            Message.log(message_type=u"d")
        g.debug = True  # output yielded messages

        with start_action(action_type=u"the-action"):
            list(g())

        assert_expected_action_tree(
            self,
            logger,
            u"the-action",
            [u"a",
             {u"confounding-factor": [u"b", u"yielded", u"c"]},
             u"d",
            ],
        )

    @capture_logging(None)
    def test_yield_inside_nested_actions(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type=u"a")
            with start_action(action_type=u"confounding-factor"):
                Message.log(message_type=u"b")
                yield None
                with start_action(action_type=u"double-confounding-factor"):
                    yield None
                    Message.log(message_type=u"c")
                Message.log(message_type=u"d")
            Message.log(message_type=u"e")
        g.debug = True  # output yielded messages

        with start_action(action_type=u"the-action"):
            list(g())

        assert_expected_action_tree(
            self,
            logger,
            u"the-action", [
                u"a",
                {u"confounding-factor": [
                    u"b",
                    u"yielded",
                    {u"double-confounding-factor": [
                        u"yielded",
                        u"c",
                    ]},
                    u"d",
                ]},
                u"e",
            ],
        )

    @capture_logging(None)
    def test_generator_and_non_generator(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type=u"a")
            yield
            with start_action(action_type=u"action-a"):
                Message.log(message_type=u"b")
                yield
                Message.log(message_type=u"c")

            Message.log(message_type=u"d")
            yield
        g.debug = True  # output yielded messages

        with start_action(action_type=u"the-action"):
            generator = g()
            next(generator)
            Message.log(message_type=u"0")
            next(generator)
            Message.log(message_type=u"1")
            next(generator)
            Message.log(message_type=u"2")
            self.assertRaises(StopIteration, lambda: next(generator))

        assert_expected_action_tree(
            self,
            logger,
            u"the-action", [
                u"a",
                u"yielded",
                u"0",
                {
                    u"action-a": [
                        u"b",
                        u"yielded",
                        u"c",
                    ],
                },
                u"1",
                u"d",
                u"yielded",
                u"2",
            ],
        )

    @capture_logging(None)
    def test_concurrent_generators(self, logger):
        @eliot_friendly_generator_function
        def g(which):
            Message.log(message_type=u"{}-a".format(which))
            with start_action(action_type=which):
                Message.log(message_type=u"{}-b".format(which))
                yield
                Message.log(message_type=u"{}-c".format(which))
            Message.log(message_type=u"{}-d".format(which))
        g.debug = True  # output yielded messages

        gens = [g(u"1"), g(u"2")]
        with start_action(action_type=u"the-action"):
            while gens:
                for g in gens[:]:
                    try:
                        next(g)
                    except StopIteration:
                        gens.remove(g)

        assert_expected_action_tree(
            self,
            logger,
            u"the-action", [
                u"1-a",
                {u"1": [
                    u"1-b",
                    u"yielded",
                    u"1-c",
                ]},
                u"2-a",
                {u"2": [
                    u"2-b",
                    u"yielded",
                    u"2-c",
                ]},
                u"1-d",
                u"2-d",
            ],
        )

    @capture_logging(None)
    def test_close_generator(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type=u"a")
            try:
                yield
                Message.log(message_type=u"b")
            finally:
                Message.log(message_type=u"c")
        g.debug = True  # output yielded messages

        with start_action(action_type=u"the-action"):
            gen = g()
            next(gen)
            gen.close()

        assert_expected_action_tree(
            self,
            logger,
            u"the-action", [
                u"a",
                u"yielded",
                u"c",
            ],
        )

    @capture_logging(None)
    def test_nested_generators(self, logger):
        @eliot_friendly_generator_function
        def g(recurse):
            with start_action(action_type=u"a-recurse={}".format(recurse)):
                Message.log(message_type=u"m-recurse={}".format(recurse))
                if recurse:
                    set(g(False))
                else:
                    yield

        g.debug = True  # output yielded messages

        with start_action(action_type=u"the-action"):
            set(g(True))

        assert_expected_action_tree(
            self,
            logger,
            u"the-action", [{
                u"a-recurse=True": [
                    u"m-recurse=True", {
                        u"a-recurse=False": [
                            u"m-recurse=False",
                            u"yielded",
                        ],
                    },
                ],
            }],
        )



import requests
from eliot import start_action, to_file
to_file(open("linkcheck.log", "w"))


def check_links(urls):
    with start_action(action_type="check_links", urls=urls):
        for url in urls:
            try:
                with start_action(action_type="download", url=url):
                    response = requests.get(url)
                    response.raise_for_status()
            except Exception as e:
                raise ValueError(str(e))

try:
    check_links(["http://eliot.readthedocs.io", "http://nosuchurl"])
except ValueError:
    print("Not all links were valid.")



"""
Tests for coroutines.

Imported into test_coroutine.py when running tests under Python 3.5 or later;
in earlier versions of Python this code is a syntax error.
"""

import asyncio
from threading import Thread
from unittest import TestCase

from ..testing import capture_logging
from ..parse import Parser
from .. import start_action
from .._action import _context_owner
from .._asyncio import AsyncioExecutionContext, use_asyncio_context


async def standalone_coro():
    """
    Log a message inside a new coroutine.
    """
    await asyncio.sleep(0.1)
    with start_action(action_type="standalone"):
        pass


async def calling_coro():
    """
    Log an action inside a coroutine, and call another coroutine.
    """
    with start_action(action_type="calling"):
        await standalone_coro()


def run_coroutines(*async_functions):
    """
    Run a coroutine until it finishes.
    """
    loop = asyncio.get_event_loop()
    futures = [asyncio.ensure_future(f()) for f in async_functions]

    async def wait_for_futures():
        for future in futures:
            await future
    loop.run_until_complete(wait_for_futures())


class CoroutineTests(TestCase):
    """
    Tests for coroutines.
    """
    def setUp(self):
        self.addCleanup(_context_owner.reset)
        use_asyncio_context()

    @capture_logging(None)
    def test_coroutine_vs_main_thread_context(self, logger):
        """
        A coroutine has a different Eliot logging context than the thread that
        runs the event loop.
        """
        with start_action(action_type="eventloop"):
            run_coroutines(standalone_coro)
        trees = Parser.parse_stream(logger.messages)
        self.assertEqual(
            sorted([(t.root().action_type, t.root().children) for t in trees]),
            [("eventloop", []), ("standalone", [])])

    @capture_logging(None)
    def test_multiple_coroutines_contexts(self, logger):
        """
        Each top-level coroutine has its own Eliot logging context.
        """
        async def waiting_coro():
            with start_action(action_type="waiting"):
                await asyncio.sleep(0.5)

        run_coroutines(waiting_coro, standalone_coro)
        trees = Parser.parse_stream(logger.messages)
        self.assertEqual(
            sorted([(t.root().action_type, t.root().children) for t in trees]),
            [("standalone", []), ("waiting", [])])

    @capture_logging(None)
    def test_await_inherits_coroutine_contexts(self, logger):
        """
        awaited coroutines inherit the logging context.
        """
        run_coroutines(calling_coro)
        [tree] = Parser.parse_stream(logger.messages)
        root = tree.root()
        [child] = root.children
        self.assertEqual(
            (root.action_type, child.action_type, child.children),
            ("calling", "standalone", []))


class ContextTests(TestCase):
    """
    Tests for coroutine support in ``eliot._action.ExecutionContext``.
    """
    def test_threadSafety(self):
        """
        Each thread gets its own execution context even when using asyncio
        contexts.
        """
        ctx = AsyncioExecutionContext()
        first = object()
        ctx.push(first)

        second = object()
        valuesInThread = []

        def inthread():
            ctx.push(second)
            valuesInThread.append(ctx.current())

        thread = Thread(target=inthread)
        thread.start()
        thread.join()
        # Neither thread was affected by the other:
        self.assertEqual(valuesInThread, [second])
        self.assertIs(ctx.current(), first)

    def test_coroutine_vs_main_thread_context(self):
        """
        A coroutine has a different Eliot context than the thread that runs the
        event loop.
        """
        ctx = AsyncioExecutionContext()
        current_context = []

        async def coro():
            current_context.append(("coro", ctx.current()))
            ctx.push("A")
            current_context.append(("coro", ctx.current()))

        ctx.push("B")
        current_context.append(("main", ctx.current()))
        run_coroutines(coro)
        current_context.append(("main", ctx.current()))
        self.assertEqual(
            current_context,
            [("main", "B"), ("coro", None), ("coro", "A"), ("main", "B")])

    def test_coroutine_vs_main_thread_context_different_thread(self):
        """
        A coroutine has a different Eliot context than the thread that runs the
        event loop, when run in different thread than the one where the context
        was created.
        """
        # Create context in one thread:
        ctx = AsyncioExecutionContext()
        current_context = []

        # Run asyncio event loop and coroutines in a different thread:
        def run():
            event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(event_loop)

            async def coro():
                current_context.append(("coro", ctx.current()))
                ctx.push("A")
                current_context.append(("coro", ctx.current()))

            ctx.push("B")
            current_context.append(("main", ctx.current()))
            run_coroutines(coro)
            current_context.append(("main", ctx.current()))

        thread = Thread(target=run)
        thread.start()
        thread.join()

        self.assertEqual(
            current_context,
            [("main", "B"), ("coro", None), ("coro", "A"), ("main", "B")])

    def test_multiple_coroutines_contexts(self):
        """
        Each top-level ("Task") coroutine has its own Eliot separate context.
        """
        ctx = AsyncioExecutionContext()
        current_context = []

        async def coro2():
            current_context.append(("coro2", ctx.current()))
            ctx.push("B")
            await asyncio.sleep(1)
            current_context.append(("coro2", ctx.current()))

        async def coro():
            current_context.append(("coro", ctx.current()))
            await asyncio.sleep(0.5)
            current_context.append(("coro", ctx.current()))
            ctx.push("A")
            current_context.append(("coro", ctx.current()))

        run_coroutines(coro, coro2)
        self.assertEqual(
            current_context,
            [("coro", None), ("coro2", None), ("coro", None),
             ("coro", "A"), ("coro2", "B")])

    def test_await_inherits_coroutine_context(self):
        """
        A sub-coroutine (scheduled with await) inherits the parent coroutine's
        context.
        """
        ctx = AsyncioExecutionContext()
        current_context = []

        async def coro2():
            current_context.append(("coro2", ctx.current()))
            ctx.push("B")
            current_context.append(("coro2", ctx.current()))

        async def coro():
            current_context.append(("coro", ctx.current()))
            ctx.push("A")
            current_context.append(("coro", ctx.current()))
            await coro2()
            current_context.append(("coro", ctx.current()))

        run_coroutines(coro)
        self.assertEqual(
            current_context,
            [("coro", None), ("coro", "A"), ("coro2", "A"),
             ("coro2", "B"), ("coro", "B")])



"""
Benchmark of message serialization.

The goal here is to mostly focus on performance of serialization, in a vaguely
realistic manner. That is, mesages are logged in context of a message with a
small number of fields.
"""

from __future__ import unicode_literals

import time

from eliot import Message, start_action, to_file

# Ensure JSON serialization is part of benchmark:
to_file(open("/dev/null"))

N = 10000


def run():
    start = time.time()
    for i in range(N):
        with start_action(action_type="my_action"):
            with start_action(action_type="my_action2"):
                Message.log(
                    message_type="my_message",
                    integer=3, string=b"abcdeft", string2="dgsjdlkgjdsl",
                    list=[1, 2, 3, 4])
    end = time.time()

    # Each iteration has 5 messages: start/end of my_action, start/end of
    # my_action2, and my_message.
    print("%.6f per message" % ((end - start) / (N * 5),))
    print("%s messages/sec" % (int(N / (end-start)),))


if __name__ == '__main__':
    run()



"""
Support for maintaining an action context across generator suspension.
"""

from __future__ import unicode_literals, absolute_import

from sys import exc_info
from functools import wraps
from contextlib import contextmanager
from weakref import WeakKeyDictionary

from ._action import _context_owner, _ExecutionContext
from . import Message


class _GeneratorContext(object):
    """Generator sub-context for C{_ExecutionContext}."""

    def __init__(self, execution_context):
        self._execution_context = execution_context
        self._contexts = WeakKeyDictionary()
        self._current_generator = None

    def init_stack(self, generator):
        """Create a new stack for the given generator."""
        stack = list(self._execution_context._get_stack())
        self._contexts[generator] = stack

    def get_stack(self):
        """Return the sub-stack for the current generator."""
        if self._current_generator is None:
            # If there is no currently active generator then we have no
            # special stack to supply.  Let the execution context figure out a
            # different answer on its own.
            return None
        # Otherwise, give back the action context stack we've been tracking
        # for the currently active generator.  It must have been previously
        # initialized (it's too late to do it now)!
        return self._contexts[self._current_generator]

    @contextmanager
    def in_generator(self, generator):
        """Context manager: set the given generator as the current generator."""
        previous_generator = self._current_generator
        try:
            self._current_generator = generator
            yield
        finally:
            self._current_generator = previous_generator


class GeneratorExecutionContext(_ExecutionContext):
    """Generator-specific C{_ExecutionContext} subclass."""

    def __init__(self):
        """This will run per-thread!"""
        _ExecutionContext.__init__(self)
        self.generator_context = _GeneratorContext(self)
        self.get_sub_context = self.generator_context.get_stack


def use_generator_context():
    """
    Make L{eliot_friendly_generator_function} work correctly.
    """
    _context_owner.set(GeneratorExecutionContext)


def _installed():
    return isinstance(_context_owner.context, GeneratorExecutionContext)


class GeneratorSupportNotEnabled(Exception):
    """
    An attempt was made to use a decorated generator without first turning on
    the generator context manager.
    """


def eliot_friendly_generator_function(original):
    """
    Decorate a generator function so that the Eliot action context is
    preserved across ``yield`` expressions.
    """
    @wraps(original)
    def wrapper(*a, **kw):
        # This isn't going to work if you don't have the generator context
        # manager installed.
        if not _installed():
            raise GeneratorSupportNotEnabled()

        # Keep track of whether the next value to deliver to the generator is
        # a non-exception or an exception.
        ok = True

        # Keep track of the next value to deliver to the generator.
        value_in = None

        # Create the generator with a call to the generator function.  This
        # happens with whatever Eliot action context happens to be active,
        # which is fine and correct and also irrelevant because no code in the
        # generator function can run until we call send or throw on it.
        gen = original(*a, **kw)

        # Initialize the per-generator Eliot action context stack to the
        # current action stack.  This might be the main stack or, if another
        # decorated generator is running, it might be the stack for that
        # generator.  Not our business.
        generator_context = _context_owner.context.generator_context
        generator_context.init_stack(gen)
        while True:
            try:
                # Whichever way we invoke the generator, we will do it
                # with the Eliot action context stack we've saved for it.
                # Then the context manager will re-save it and restore the
                # "outside" stack for us.
                #
                # Regarding the support of Twisted's inlineCallbacks-like
                # functionality (see eliot.twisted.inline_callbacks):
                #
                # The invocation may raise the inlineCallbacks internal
                # control flow exception _DefGen_Return.  It is not wrong to
                # just let that propagate upwards here but inlineCallbacks
                # does think it is wrong.  The behavior triggers a
                # DeprecationWarning to try to get us to fix our code.  We
                # could explicitly handle and re-raise the _DefGen_Return but
                # only at the expense of depending on a private Twisted API.
                # For now, I'm opting to try to encourage Twisted to fix the
                # situation (or at least not worsen it):
                # https://twistedmatrix.com/trac/ticket/9590
                #
                # Alternatively, _DefGen_Return is only required on Python 2.
                # When Python 2 support is dropped, this concern can be
                # eliminated by always using `return value` instead of
                # `returnValue(value)` (and adding the necessary logic to the
                # StopIteration handler below).
                with generator_context.in_generator(gen):
                    if ok:
                        value_out = gen.send(value_in)
                    else:
                        value_out = gen.throw(*value_in)
                    # We have obtained a value from the generator.  In
                    # giving it to us, it has given up control.  Note this
                    # fact here.  Importantly, this is within the
                    # generator's action context so that we get a good
                    # indication of where the yield occurred.
                    #
                    # This is noisy, enable only for debugging:
                    if wrapper.debug:
                        Message.log(message_type=u"yielded")
            except StopIteration:
                # When the generator raises this, it is signaling
                # completion.  Leave the loop.
                break
            else:
                try:
                    # Pass the generator's result along to whoever is
                    # driving.  Capture the result as the next value to
                    # send inward.
                    value_in = yield value_out
                except:
                    # Or capture the exception if that's the flavor of the
                    # next value.  This could possibly include GeneratorExit
                    # which turns out to be just fine because throwing it into
                    # the inner generator effectively propagates the close
                    # (and with the right context!) just as you would want.
                    # True, the GeneratorExit does get re-throwing out of the
                    # gen.throw call and hits _the_generator_context's
                    # contextmanager.  But @contextmanager extremely
                    # conveniently eats it for us!  Thanks, @contextmanager!
                    ok = False
                    value_in = exc_info()
                else:
                    ok = True

    wrapper.debug = False
    return wrapper



"""Support for Eliot tracing with Dask computations."""

from pyrsistent import PClass, field

from dask import compute, optimize
from dask.core import toposort, get_dependencies
from . import start_action, current_action, Action, Message


class _RunWithEliotContext(PClass):
    """
    Run a callable within an Eliot context.

    @ivar task_id: The serialized Eliot task ID.
    @ivar func: The function that Dask wants to run.
    @ivar key: The key in the Dask graph.
    @ivar dependencies: The keys in the Dask graph this depends on.
    """
    task_id = field(type=str)
    func = field()  # callable
    key = field(type=str)
    dependencies = field()

    # Pretend to be underlying callable for purposes of equality; necessary for
    # optimizer to be happy:

    def __eq__(self, other):
        return self.func == other

    def __ne__(self, other):
        return self.func != other

    def __hash__(self):
        return hash(self.func)

    def __call__(self, *args, **kwargs):
        with Action.continue_task(task_id=self.task_id):
            Message.log(
                message_type="dask:task",
                key=self.key,
                dependencies=self.dependencies
            )
            return self.func(*args, **kwargs)


def compute_with_trace(*args):
    """Do Dask compute(), but with added Eliot tracing.

    Dask is a graph of tasks, but Eliot logs trees.  So we need to emulate a
    graph using a tree.  We do this by making Eliot action for each task, but
    having it list the tasks it depends on.

    We use the following algorithm:

        1. Create a top-level action.

        2. For each entry in the dask graph, create a child with
           serialize_task_id.  Do this in likely order of execution, so that
           if B depends on A the task level of B is higher than the task Ievel
           of A.

        3. Replace each function with a wrapper that uses the corresponding
           task ID (with Action.continue_task), and while it's at it also
           records which other things this function depends on.

    Known issues:

        1. Retries will confuse Eliot.  Probably need different
           distributed-tree mechanism within Eliot to solve that.
    """
    # 1. Create top-level Eliot Action:
    with start_action(action_type="dask:compute"):
        # In order to reduce logging verbosity, add logging to the already
        # optimized graph:
        optimized = optimize(*args, optimizations=[_add_logging])
        return compute(*optimized, optimize_graph=False)


def _add_logging(dsk, ignore=None):
    """
    Add logging to a Dask graph.

    @param dsk: The Dask graph.

    @return: New Dask graph.
    """
    ctx = current_action()
    result = {}

    # Use topological sort to ensure Eliot actions are in logical order of
    # execution in Dask:
    keys = toposort(dsk)

    # Give each key a string name. Some keys are just aliases to other
    # keys, so make sure we have underlying key available. Later on might
    # want to shorten them as well.
    def simplify(k):
        if isinstance(k, str):
            return k
        return "-".join(str(o) for o in k)

    key_names = {}
    for key in keys:
        value = dsk[key]
        if not callable(value) and value in keys:
            # It's an alias for another key:
            key_names[key] = key_names[value]
        else:
            key_names[key] = simplify(key)

    # 2. Create Eliot child Actions for each key, in topological order:
    key_to_action_id = {
        key: str(ctx.serialize_task_id(), "utf-8")
        for key in keys
    }

    # 3. Replace function with wrapper that logs appropriate Action:
    for key in keys:
        func = dsk[key][0]
        args = dsk[key][1:]
        if not callable(func):
            # This key is just an alias for another key, no need to add
            # logging:
            result[key] = dsk[key]
            continue
        wrapped_func = _RunWithEliotContext(
            task_id=key_to_action_id[key],
            func=func,
            key=key_names[key],
            dependencies=[key_names[k] for k in get_dependencies(dsk, key)],
        )
        result[key] = (wrapped_func, ) + tuple(args)

    assert result.keys() == dsk.keys()
    return result


__all__ = ["compute_with_trace"]



"""
Support for asyncio coroutines.
"""

try:
    from asyncio import Task
except ImportError:
    Task = None  # Python 2
from weakref import WeakKeyDictionary

from ._action import _ExecutionContext


class AsyncioSubContext:
    """
    Per-Task context, allowing different coroutines to have different logging
    context.

    This will be attached to threading.local object, so no need to worry about
    thread-safety.
    """
    def __init__(self):
        self._per_task = WeakKeyDictionary()

    def get_stack(self):
        """
        Get the stack for the current Task, or None if there is no Task.
        """
        try:
            task = Task.current_task()
        except RuntimeError:
            # No loop for this thread:
            task = None
        if task is None:
            return None
        if task not in self._per_task:
            self._per_task[task] = []
        return self._per_task[task]


class AsyncioExecutionContext(_ExecutionContext):
    """ExecutionContext that supports asyncio sub-contexts."""

    def __init__(self):
        _ExecutionContext.__init__(self)
        self.get_sub_context = AsyncioSubContext().get_stack


def use_asyncio_context():
    """
    Use a logging context that is tied to the current asyncio coroutine.

    Call this first thing, before doing any other logging.

    Does not currently support event loops other than asyncio.
    """
    from ._action import _context_owner
    _context_owner.set(AsyncioExecutionContext)
